const char *help = "\
SVMTorch III (c) Trebolloc & Co 2002\n\
\n\
This program will train a SVM for classification or regression\n\
with a gaussian kernel\n";

#include "MatDataSet.h"
#include "TwoClassFormat.h"
#include "ClassMeasurer.h"
#include "MSEMeasurer.h"
#include "QCTrainer.h"
#include "CmdLine.h"
#include "Random.h"
#include "SVMRegression.h"
#include "SVMClassification.h"
#include "KFold.h"
#include "DiskXFile.h"
#include "ClassFormatDataSet.h"

using namespace Torch;

int main(int argc, char **argv)
{
  char *file;
  real c_cst, stdv, eps_regression;
  bool regression;
  real accuracy, cache_size;
  int iter_shrink, k_fold;
  int the_seed;
  int max_load;
  char *dir_name;
  char *model_file;
  int max_load_test;
  bool binary_mode;

  Allocator *allocator = new Allocator;

  //=================== The command-line ==========================

  // Construct the command line
  CmdLine cmd;

  // Put the help line at the beginning
  cmd.info(help);

  // Train mode
  cmd.addText("\nArguments:");
  cmd.addSCmdArg("file", &file, "the train file");

  cmd.addText("\nModel Options:");
  cmd.addRCmdOption("-c", &c_cst, 100., "trade off cst between error/margin");
  cmd.addRCmdOption("-std", &stdv, 10., "the std parameter in the gaussian kernel", true);
  cmd.addBCmdOption("-rm", &regression, false, "regression mode", true);
  cmd.addRCmdOption("-eps", &eps_regression, 0.7, "eps tube in regression");

  cmd.addText("\nLearning Options:");
  cmd.addRCmdOption("-e", &accuracy, 0.01, "end accuracy");
  cmd.addRCmdOption("-m", &cache_size, 50., "cache size in Mo");
  cmd.addICmdOption("-h", &iter_shrink, 100, "minimal number of iterations before shrinking");
  cmd.addICmdOption("-kfold", &k_fold, -1, "number of folds, if you want to do cross-validation");

  cmd.addText("\nMisc Options:");
  cmd.addICmdOption("-seed", &the_seed, -1, "the random seed");
  cmd.addICmdOption("-load", &max_load, -1, "max number of examples to load for train", true);
  cmd.addSCmdOption("-dir", &dir_name, ".", "directory to save measures");
  cmd.addSCmdOption("-save", &model_file, "", "the model file");
  cmd.addBCmdOption("-bin", &binary_mode, false, "binary mode for files");

  // Test mode
  cmd.addMasterSwitch("--test");
  cmd.addText("\nArguments:");
  cmd.addSCmdArg("model", &model_file, "the model file");
  cmd.addSCmdArg("file", &file, "the test file");

  cmd.addText("\nMisc Options:");
  cmd.addICmdOption("-load", &max_load_test, -1, "max number of examples to load for train");
  cmd.addSCmdOption("-dir", &dir_name, ".", "directory to save measures");
  cmd.addBCmdOption("-bin", &binary_mode, false, "binary mode for files");

  // Read the command line
  int mode = cmd.read(argc, argv);

  DiskXFile *model = NULL;
  if(mode == 1)
  {
    model = new(allocator) DiskXFile(model_file, "r");
    cmd.loadXFile(model);
  }

  // If the user didn't give any random seed,
  // generate a random random seed...
  if(mode == 0)
  {
    if(the_seed == -1)
      Random::seed();
    else
      Random::manualSeed((long)the_seed);
  }
  cmd.setWorkingDirectory(dir_name);

  //=================== Create the SVM... =========================
  SVM *svm = NULL;
  GaussianKernel kernel(1./(stdv*stdv));

  if(regression)
    svm = new(allocator) SVMRegression(&kernel);
  else
    svm = new(allocator) SVMClassification(&kernel);

  if(mode == 0)
  {
    svm->setROption("C", c_cst);
    svm->setROption("cache size", cache_size);
    if(regression)
      svm->setROption("eps regression", eps_regression);
  }
  //=================== DataSets & Measurers... ===================

  // Create the training dataset (normalize inputs)
  DataSet *data = NULL;
  if(regression)
    data = new(allocator) MatDataSet(file, -1, 1, false, max_load, binary_mode);
  else
  {
    MatDataSet *mat_data = new(allocator) MatDataSet(file, -1, 1, false, max_load, binary_mode);
    Sequence *class_labels = new(allocator) Sequence(2, 1);
    class_labels->frames[0][0] = -1;
    class_labels->frames[0][1] = 1;
    data = new(allocator) ClassFormatDataSet(mat_data, class_labels);
  }

  // The list of measurers...
  MeasurerList measurers;
  if(mode == 0)
  {
    if(k_fold > 0)
    {
      if(regression)
      {
        MSEMeasurer *mse_meas = new(allocator) MSEMeasurer(svm->outputs, data, cmd.getXFile("the_mse_err"));
        measurers.addNode(mse_meas);
      }
      else
      {
        TwoClassFormat *class_format = new(allocator) TwoClassFormat(data);
        ClassMeasurer *class_meas = new(allocator) ClassMeasurer(svm->outputs, data, class_format, cmd.getXFile("the_class_err"));
        measurers.addNode(class_meas);
      }
    }
  }
  else
  {
    // Reload
    svm->loadXFile(model);

    // Test class measurer
    if(regression)
    {
      MSEMeasurer *test_mse_meas = new(allocator) MSEMeasurer(svm->outputs, data, cmd.getXFile("the_test_mse_err"));
      measurers.addNode(test_mse_meas);
    }
    else
    {
      TwoClassFormat *class_format = new(allocator) TwoClassFormat(data);
      ClassMeasurer *test_class_meas = new(allocator) ClassMeasurer(svm->outputs, data, class_format, cmd.getXFile("the_test_class_err"));
      measurers.addNode(test_class_meas);
    }
  }
  
  //=================== The Trainer ===============================

  QCTrainer trainer(svm);

  //=================== Let's go... ===============================

  if(mode == 0)
  {
    if(k_fold <= 0)
    {
      trainer.train(data, NULL);
      message("%d SV with %d at bounds", svm->n_support_vectors, svm->n_support_vectors_bound);
      if(strcmp(model_file, ""))
      {
        DiskXFile model_(model_file, "w");
        cmd.saveXFile(&model_);
        data->saveXFile(&model_);
        svm->saveXFile(&model_);
      }
    }
    else
    {
      KFold k(&trainer, k_fold);
      k.crossValidate(data, NULL, &measurers);
    }
  }
  else
    trainer.test(&measurers);

  delete allocator;

  return(0);
}
